# Copyright 2007, Sjoerd de Vries
# Copyright 2007, Sjoerd de Vries
# This file is part of the Spyder main program (a.k.a the Spyder kernel)
# For licensing information, see LICENSE.txt 

import re, sys, safe_eval
from copy import copy

reservedtypes = ("Spyder", "Type", "Delete", "Exception", "None", "True", "False")
def validvar(s):
  if s.replace("_", "x").isalnum() == False: return False
  if s[0].isupper() == False: return False
  if len(s) > 1 and s == s.upper(): return False
  if s.endswith("Array") or s.endswith("Error") or s.endswith("Exit") or s in reservedtypes: return False
  return True

quotematch = re.compile(r'(([\"\']).*?\2)')
triplequotematch = re.compile(r'(\"\"\"[\w\Wn]*?\"\"\")')
curlymatch = re.compile(r'{[^{}]*?}')

def getlines(s):
  pos = 0
  s0 = ""
  for pp in triplequotematch.finditer(s):
    s0 += s[pos:pp.start()] + "&" * (pp.end()-pp.start())
    pos = pp.end()
  s0 += s[pos:]

  p = quotematch.finditer(s0)
  pos = 0
  mask0 = ""
  for pp in p:
      mask0 += s[pos:pp.start()] + "*" * (pp.end()-pp.start())
      pos = pp.end()
  mask0 += s[pos:]

  while 1:
    p = curlymatch.finditer(mask0)
    pos = 0
    mask = ""
    for pp in p:
      mask += mask0[pos:pp.start()] + "!" * (pp.end()-pp.start())
      pos = pp.end()
    mask += mask0[pos:]
    if pos == 0: break
    mask0 = mask

  
  pos = 0
  mask = ""
  for pp in triplequotematch.finditer(mask0):
    mask += mask0[pos:pp.start()] + "!" * (pp.end()-pp.start())
    pos = pp.end()
  mask += mask0[pos:]
  
  lines0 = mask.split("\n")
  lines = []
  pos = 0
  for l in lines0:
    pos2 = pos + len(l)
    lines.append(s[pos:pos2])
    pos = pos2 + len("\n")
  return lines

def getblock(s):
  
  pos = 0
  s0 = ""
  for pp in triplequotematch.finditer(s):
    s0 += s[pos:pp.start()] + "&" * (pp.end()-pp.start())
    pos = pp.end()
  s0 += s[pos:]
  
  p = quotematch.finditer(s0)
  pos = 0
  mask0 = ""
  for pp in p:
      mask0 += s[pos:pp.start()] + "*" * (pp.end()-pp.start())
      pos = pp.end()
  mask0 += s[pos:]

  preblock = s
  postblock = ""
  blocks = []
  while 1:
    p = curlymatch.finditer(mask0)
    pos = 0
    mask = ""
    blocks0 = []
    for pp in p:
        blocks0.append(s[pp.start():pp.end()])
        preblock = s[pos:pp.start()]
        mask += preblock + "!" * (pp.end()-pp.start())
        pos = pp.end()
    mask += mask0[pos:]
    if pos == 0:
      break
    else: postblock = s[pos:]
    blocks = blocks0
    mask0 = mask
  block = None
  if len(blocks) == 1: 
    block = blocks[0][1:-1]
  if len(blocks) > 1:
    raise Exception("compile error: invalid statement\n%s\nStatement must contain a single {} block" % s)
    if len(postblock.strip()) != 0:
      raise Exception("compile error: invalid statement\n%s\nStatement must be empty after {} block" % s)
  preblock = preblock.strip()
  
  blockcomment = ""
  
  p = quotematch.finditer(preblock)
  pos = 0
  mask = ""
  for pp in p:
      mask += preblock[pos:pp.start()] + "*" * (pp.end()-pp.start())
      pos = pp.end()
  mask += preblock[pos:]  
  comment = mask.find("#")
  if comment > -1:     
    blockcomment = preblock[comment+1:].strip('\n') + '\n'
    preblock = preblock[:comment]
  
  if block != None:
    currblock = block
    while 1:
      len0 = len(currblock)
      currblock = currblock.lstrip().lstrip("\n")
      if len(currblock) == len0: break
      
    pp = triplequotematch.search(currblock)
    if pp != None and pp.start() == 0:
      blockcomment += currblock[pp.start()+len('"""'):pp.end()-len('"""')]
            
  name = None
  title = None
  if len(preblock) > 0:
    name = preblock.split()[0]
    title = preblock[len(name):].strip()
    
  return name,title,block, blockcomment

def typedefparse(block, textblocks, macros):  
  filtersource = ""
  commentblock = ""
    
  lines = getlines(block)
  defmode = False
  currindent = 0
  prelines = 0
  memberblock = None
  while len(lines) > 0:
    if prelines > 0: prelines -= 1
    l = lines[0].strip()
    l2 = lines[0].replace('\t', "  ")
    lines = lines[1:]
    if len(l) == 0: continue
    if l.startswith("##"):
      l = l[2:].lstrip()
      ll = l.split()
      name = ll[0]
      if name not in textblocks: textblocks[name] = ""
      textblocks[name] += l[len(name)+1:] + '\n'
      continue
    if l.find("#") == 0: 
      spacing = l2[:len(l2)-len(l)]
      filtersource += spacing + l + '\n'
      continue
    if defmode == False and l.startswith('"""'):
      spacing = l2[:len(l2)-len(l)]
      l = l[3:]
      l = '\n' + l[:l.index('"""')]
      filtersource += spacing + "#!#!#!\n" + spacing + l.strip('\n').strip() + '\n' + spacing + "#!#!#!\n"
      if memberblock: textblocks["members"][-1][2] += l[len('\n'):]
      elif memberblock == None: commentblock += l[len('\n'):]
      continue
    if defmode == True:
      indent = len(l2) - len(l2.lstrip())
      if indent == currindent:
        defmode = False
    if defmode == False:
      if l2.lstrip().startswith("def "):
        currindent = len(l2) - len(l2.lstrip())
        defmode = True
        if prelines == 0: filtersource += l2 + '\n'
    if defmode == True or l2.lstrip().startswith("@"):      
      if "__endclass" not in textblocks: textblocks["__endclass"] = ""
      textblocks["__endclass"] += l2 + '\n'
      continue
    if prelines == 0: filtersource += l2 + '\n'
    
    name,title,block,blockcomment = getblock(l)
    memberblock = False
    if block != None:
      if title != "" and title != None:
        raise Exception("compile error: unimethod statement must be <name> {...}\n%s" % (l))
      if name not in textblocks: textblocks[name] = ""
      bb = block.split('\n')
      spaces = -1
      for l in bb:
        if len(l.strip()) == 0: continue
        if spaces == -1:
          spaces = len(l) - len(l.lstrip())
        textblocks[name] += l.rstrip('\n')[spaces:] + '\n'
    else:
      triggermacro = False
      for m in macros:
        mm = m(name, title)
        if mm != None:
          triggermacro = True
          oldlen = len(lines)
          lines = getlines(mm) + lines
          prelines += len(lines) - oldlen
          break
      if triggermacro == False:
        if name == "Delete":
          if title in textblocks:
            textblocks.pop(title)  
          textblocks["members"] = [a for a in textblocks["members"] if a[1].split("=")[0].rstrip() != title]
          textblocks["members"].append([name,title, blockcomment])
        else:
          memberblock = True 
          textblocks["members"].append([name,title, blockcomment])
          

  #textblocks["members"] = oldtextblocks["members"] + textblocks["members"]
  #checkcounter = len(textblocks["members"])
  """
  for (membertype, title) in textblocks["members"]:
    if membertype == "Delete":
      if title in oldtextblocks:
        oldtextblocks.pop(title)
 
  for v in oldtextblocks:
    if v != "members":
      if v not in textblocks: textblocks[v] = ""
      textblocks[v] += oldtextblocks[v]
  """
  members0 = {}
  memberlist = set()
  counter = 0
  for (membertype, title, blockcomment) in textblocks["members"]:
    if membertype == "Delete": 
      members0[len(members0)] = [membertype, title, blockcomment, len(members0)]
      continue
    """
    if membertype == "Delete":
      if title in textblocks:
        if counter <= checkcounter: continue        
        textblocks.pop(title)
      if title in members0:
        if "form" not in textblocks: textblocks["form"] = ""
        textblocks["form"] += "DELETE %s\n" % title                
        members0[title][0] = None
      continue
    """
    membername = title.split("=")[0].rstrip()
    if membername.replace("_","x").isalnum() == False:
      raise Exception("compile error: invalid membername %s" % (membername))
    if membername in members0 and members0[membername][0] != None:
      raise Exception("compile error: duplicate membername %s" % membername)
    if membername in members0: members0[membername][0] = membertype
    else: members0[membername] = [membertype, title, blockcomment, len(members0)]
  members00 = {}
  for m in members0:
    if members0[m][0] != None:
      members00[members0[m][3]] = (members0[m][0], members0[m][1], members0[m][2])
  memnrs = copy(members00.keys())
  memnrs.sort()
  textblocks["members"] = []
  for nr in memnrs: textblocks["members"].append(members00[nr])
  #print "MEMBERS", textblocks["members"]
  return filtersource, commentblock

def typedef(l, title, block):
  global neededtypes
  noerror = True
  if len(title) == 0 or block == None:
    raise Exception("compile error: invalid type definition\n%s\nType definition must have a proper title" % l)
  t = title.split(":")
  if len(t) > 2:
    raise Exception("compile error: invalid type definition\n%s\nInvalid title: %s" % (l, title))
  typename = t[0]
  if  validvar(typename) == False:
    raise Exception("compile error: invalid type definition\n%s\nInvalid title: %s" % (l, title))  
  if typename in reservedtypes:
    raise Exception("compile error: reserved type name %s" % (typename))
  if typename in definedtypes and len(neededtypes) > 0:
    return ""
  if typename in inheritedtypes:
    raise Exception("compile error: type name %s has already been defined" % (typename))
  oldtextblocks = {}
  oldtextblocks["members"] = []
  inherited = None
  if len(t) == 2:
    inherited = t[1]
    if  validvar(inherited) == False:
      raise Exception("compile error: invalid type definition\n%s\nInvalid title: %s" % (l, title))
    if inherited.endswith("Array"):
      raise Exception("compile error: invalid type definition\n%s\nSpyder inheritance from Spyder array is not implemented: %s" % (l, title))
    parenttype = None
    if inherited in inheritedtypes:
      parenttype = inheritedtypes[inherited]
    elif inherited in definedtypes:
      parenttype = definedtypes[inherited]
    if parenttype == None:
      neededtypes.add(inherited)
      noerror = False      
    else:
      for v in parenttype:
        if not v.startswith("constructor_"):
          oldtextblocks[v] = copy(parenttype[v])

  textblocks = oldtextblocks
        
  filtersource = "Spyder-generated class\n\n module %s\n file %s\n\n" % (currmodule, currfile)
  
  filtersource0, commentblock = typedefparse(block, textblocks, macros) 
  #textblocks gets updated...
  
  filtersource += commentblock + '\n'
  
  filtersource += 70 * "-" + "\nSpyder definition:\n\n"          
  if inherited != None:
    filtersource += "Type %s(%s) {\n" % (typename, inherited)
  else: 
    filtersource += "Type %s {\n" % (typename)  

  filtersource += filtersource0
  textblocks["members"] = [a for a in textblocks["members"] if a[0] != "Delete"]
  filtersource += "}\n"
  filtersource += '\n' + 70 * "-" + '\n'
  
  code = ""  
  unimethods0 = copy(unimethodlist)
  unimethods0.remove("__class")
  unimethods0.remove("__endclass")
  endclass = {}
  allmethods = list(["__class",] + unimethods0)
  for u in allmethods:
    block = None
    if u in textblocks: block = textblocks[u]
    unimethodcode, endclasscode = unimethods[u](typename, inherited, filtersource, textblocks["members"], block)
    endclass[u] = endclasscode
    if unimethodcode != None: code += unimethodcode
  block = None
  if "__endclass" in textblocks: block = textblocks["__endclass"]    
  endclasscode = unimethods["__endclass"](typename, inherited, filtersource, textblocks["members"], block, endclass)
  if endclasscode != None: code += endclasscode
  if noerror == True:
    #safe_eval.safe_eval(code, True) DISABLED! unimethods must check!
    if "__endclass" in textblocks:
      try:
        testcode = "class bogus:\n" + textblocks["__endclass"]
        safe_eval.safe_eval(testcode)
      except:
        #dump = open("dump.py", "w")
        #dump.write(testcode)
        #dump.close()        
        raise
    if inherited != None: textblocks["__parent"] = inherited 
    definedtypes[typename] = textblocks
    Spyder.core.__defined[typename] = textblocks
    return code
  else:
    return ""

def statement(l0):
  global result, nondefinitions
  l = l0.strip()
  if len(l) == 0: return
  if l[0] == "#": return

  name,title,block,blockcomment = getblock(l)
  
  if name == "Type":
    code = typedef(l,title, block) + '\n'
    result += code
    return
  if len(neededtypes) > 0: return
  found = False
  for statement in statements:
    if name == statement:
      sup = set()
      r, sup = statements[statement](l, title, block)
      for suptyp in sup:
        while suptyp.endswith("Array"):
          suptyp=suptyp[:-len("Array")]
        if suptyp != "None": supportedtypes.add(suptyp)
      result += r
      #nondefinitions += r #DISABLED: statements must check for safe_eval
      found = True
      break
  if found == False:
      result += l0 + '\n'
      nondefinitions += l0 + '\n'
    
def Spydercompile(spyder, s, unimethodlist0, unimethods0, macros0, statements0, inheritedtypes0, currmodule0, currfile0):
  global unimethods
  global unimethodlist
  global macros
  global statements
  global inheritedtypes
  global Spyder
  global nondefinitions
  global result, neededtypes, definitions, definedtypes, supportedtypes
  global currmodule, currfile
  currmodule, currfile = currmodule0, currfile0
  unimethods = unimethods0
  unimethodlist = unimethodlist0
  macros = macros0
  statements = statements0
  inheritedtypes = inheritedtypes0
  lines = getlines(s)
  Spyder = spyder
  result = ""
  definitions = ""
  nondefinitions = ""
  neededtypes = set()
  definedtypes = {}
  supportedtypes = set()
  for l in lines:
    statement(l)
  if len(neededtypes) > 0: result = None
  safe_eval.safe_eval(nondefinitions)
  return result, neededtypes, definedtypes, supportedtypes

