# Copyright 2007, Sjoerd de Vries
# This file is part of the Spyder main program (a.k.a the Spyder kernel)
# For licensing information, see LICENSE.txt 

__switches = set()
__tempdir = None
__args = []
__spyderobjects = {}
__state = None

try:
  import sys, os
  __spyderdir = os.path.abspath(os.path.split(__file__)[0])
  sys.path.insert(0, __spyderdir)  
  import system as __system, Spydercompile as __Spydercompile
  reload(__system)
  reload(__Spydercompile)
except ImportError:
  raise ImportError("Spyder could not locate its core functions. Your Spyder install seems to be broken")

import imp as __imp, functools as __functools
__namespace = __functools.partial(__imp.new_module, "Spydernamespace")

def __register_spyderobject(spyderstr, obj):
  global __spyderobjects
  __spyderobjects[spyderstr] = obj

def __retrieve_spyderobject(spyderstr):
  try:
    return __spyderobjects[spyderstr]
  except:
    return None

def __retrieve_all_spyderobjects():
  return __spyderobjects

def __exception():
  return str(sys.exc_info()[1])

def __constructor(__funcname__, __class__, __defaultconstructor__, *args, **kargs):    
  __class__.__constructor__ = __funcname__
  ret = __class__(*args, **kargs)
  __class__.__constructor__ = __defaultconstructor__
  return ret

try:
  __systemfile = open(os.path.join(__spyderdir,"system.fly"))
except IOError:
  raise IOError("Cannot locate system.fly. Did you install Spyder properly?")

__systemlines = __systemfile.readlines()
__systemlines_backup = [s for s in __systemlines]
__systemfile.close()

def __do_extend_array(tt):
  arraycounter = 0
  ret = False
  while tt.endswith("Array"):
    arraycounter += 1
    tt = tt[:-len("Array")]
  if tt not in __state.arraycount: __state.arraycount[tt] = 0
  if arraycounter > __state.arraycount[tt]: 
    ret = True      
    __state.arraycount[tt] = arraycounter
  return ret

def __load_namespaces(namespaces, arraytypes = []):
  olddir = os.getcwd()
  currnrs = [v[0] for v in __state.allnamespaces.values()]
  newnrs = [v[1] for v in namespaces if v[1] not in currnrs]+[999]
  currmaxnr = max(currnrs+[-1])
  minnr = min(newnrs)
  extend_array = False
    
  for nam, nr, namespace in namespaces:
    __state.allnamespaces[nam] = (nr, namespace)      
    for tt in namespace.calledtypes:
      if __do_extend_array(tt): extend_array = True
  for tt in arraytypes: 
    if __do_extend_array(tt): extend_array = True
    
  if minnr > currmaxnr and currmaxnr > -1 and not extend_array:      
    allnam = [v for v in namespaces if v[1] in newnrs]
    allnam.sort(__sortfunc)
    for nam, nr, namespace in allnam:
      a = [nam, namespace.codefiles, namespace.paths, namespace.imports, namespace.exports, namespace.defined]        
      __system.changetempdir(__tempdir)
      code = __system.import_namespace(*a)
      os.chdir(olddir)
      exec(code, __shell)
  else:      
    __run()

def __load_spydertype(spydertype):
  st0 = spydertype
  while st0.endswith("Array"): st0 = st0[:-len("Array")]
  return __load_namespaces([[v[0], v[1], __state.namespaces[v[0]]] for v in __state.namespacemapping[st0]], [spydertype])

def __load_namespace(nam):
  nr = __system.namespacelist.index(nam)
  namespace = __state.namespaces[nam]
  for nam2 in namespace.called:
    if nam2 != nam: __load_namespace(nam2)
  return __load_namespaces([(nam,nr,namespace)])

def __addsystemtypes(code, definedtypes, currstate):
  arraycount = currstate.arraycount
  namespacemapping = currstate.namespacemapping
  allnamespaces = currstate.allnamespaces
  for t in definedtypes:
    if t in system.inheritedtypes:
      raise Exception("Compile error: duplicate definition of type %s" % (t))
    system.inheritedtypes[t] = definedtypes[t]
  calledtypes = []
  calledtypes += __system.get_called_types(code)
  calledtypes = set(calledtypes)
  for t in calledtypes:
    arraycounter = 0
    while t.endswith("Array"):
      arraycounter += 1
      t = t[:-len("Array")]
    if t not in arraycount: arraycount[t] = 0
    if arraycounter > arraycount[t]: arraycount[t] = arraycounter
    if t in definedtypes: continue
    if t in currstate.commandlinetypes: continue
    if t not in namespacemapping or len(namespacemapping[t]) == 0:      
      raise Exception("Execution error: the system could not provide the following type:\n " + t)
    namcalledtypes = set()
    for nam, nr in namespacemapping[t]:
      namespace = currstate.namespaces[nam]
      namcalledtypes = namcalledtypes.union(namespace.calledtypes)
      if nam not in allnamespaces and len(namespace.codefiles) > 0:
        allnamespaces[nam] = (nr, namespace)
      
    for tt in namcalledtypes:
      arraycounter = 0
      while tt.endswith("Array"):
        arraycounter += 1
        tt = tt[:-len("Array")]
      if tt not in arraycount: arraycount[tt] = 0
      if arraycounter > arraycount[tt]: arraycount[tt] = arraycounter
    currstate.arraycount = arraycount

def __reset():
  global __state, __shell, __switches, __args,__systemlines, __toplevelfiles
  __systemlines = [s for s in __systemlines_backup]
  __switches = set()
  __args = []
  core = __namespace()
  __system.Spyder.core = core
  core.namespace = __namespace
  core.defineunimethod = __system.defineunimethod
  core.definemacro = __system.definemacro
  core.definestatement = __system.definestatement
  core.__currpaths = []
  core.__arraycount = {}
  core.__types__ = {}      
  core.__defined = {}
  __system.reset()
  

def init(*args, **kargs):  
  if (args,kargs) == ((None,), {}): return __init(__switches, None, __args)
  
  switchonly = True
  if len(kargs): switchonly = False
  else:
    for v in args:
      if not isinstance(v, str): switchonly = False
  if switchonly:
    return __init(args)    
  else:
    arguments = ["switches", "tempdir", "args"]
    iniargs = {}
    for anr, a in enumerate(arguments):
      d1 = len(args) > anr
      d2 = a in kargs
      if d1 and d2: raise Exception("Spyder.init: duplicate definition of \"%s\"" % a)
      if d1: iniargs[a] = args[anr]
      elif d2: iniargs[a] = kargs[a]
      else: iniargs[a] = None
    if iniargs["args"] == None: iniargs["args"] = []
    return __init(**iniargs)
      
def __init(switches=set(), tempdir=None, args=[]):  
  global __state, __shell, __switches, __corecode, __systemlines, __toplevelfiles, __tempdir
  __reset()
  __switches = set(switches)
  if tempdir != None: __tempdir = str(tempdir)
  __state = __namespace()  
  __toplevelfiles = []
  olddir = os.getcwd()
  filteredargs = []
  for arg in args: 
    arg = arg.replace("\\","\\\\")
    __systemlines.append(arg)
  if __tempdir != None:
    if not os.path.exists(__tempdir):
      raise Exception("TEMPDIR %s does not exist" % __tempdir)     
    __system.changetempdir(__tempdir)      
  __corecode, __state.namespacemapping, __state.namespaces, __toplevelfiles, newtempdir = __system.systemcompile(__systemlines, __switches,__spyderdir,(__tempdir!=None))
  if newtempdir != None: __tempdir = newtempdir
  __state.allnamespaces = {}
  __state.arraycount = {}
  __state.commandlinetypes = []
  __reset()
  shellspace = __namespace()
  shellspace.core = __system.Spyder.core
  shellspace.getvar = getvar
  shellspace.getvars = getvars
  __shell = {'Spyder':shellspace}
  
  os.chdir(olddir)
  exec(__corecode, __shell)
  for typ in __state.namespacemapping.keys():
    for n in range(5): #0-5x array
      name = typ + n * "Array"
      globals()[name] = __wrapper(name)
  for mod in __state.namespaces.keys():
    globals()[mod] = __module_wrapper(mod)
  
  shellspace = __namespace()
  shellspace.core = __system.Spyder.core
  shellspace.getvar = getvar
  shellspace.getvars = getvars  
  __shell = {'Spyder':shellspace}
  
#the temporary dictionary is reset after every invocation of the conversion engine

__spyder_temp = {}
def __set_temp(key, value):
  global __spyder_temp
  __spyder_temp[key] = value
  
def __get_temp(key):
  global __spyder_temp
  return __spyder_temp.get(key,None)

def __clear_temp():
  global __spyder_temp
  __spyder_temp = {}
  
def __exception():
  return str(sys.exc_info()[1])

def __sortfunc(x,y):
  return cmp(x[1], y[1])

def __run():
  olddir = os.getcwd()
  global __args
  __args = []
  codefiles = []

  code = ""
  compiledfilenames = set()
  for fil in __toplevelfiles:
    ok = True
    try:  
      f = open(fil)
      fcode = f.read().replace('\r\n','\n')
      f.close()
    except:
      ok = False
    if fil.upper().endswith("SPY"):
      if not ok: raise
      result, neededtypes, definedtypes, supportedtypes = __Spydercompile.Spydercompile(
        __system.Spyder,
        fcode,
        __system.unimethodlist,
        __system.unimethods,
        __system.macros,
        __system.statements,
        __system.inheritedtypes,
        "<Main program>",
        fil,
      )      
      if result == None:
        raise Exception("Compile error: the system could not provide the following types:\n " + str(list(neededtypes)))
      __state.commandlinetypes += definedtypes.keys()
      result = result.strip('\n') + '\n'
      __addsystemtypes(result,definedtypes,__state)
      nrlines = len(result.rstrip().splitlines())
      compiledfilename = fil[:-4]
      if __tempdir != None: 
        compiledfilename = os.path.basename(compiledfilename)
        if compiledfilename in compiledfilenames:
          raise Exception("Duplicate file %s" % compiledfilename)
        compiledfilenames.add(compiledfilename)
      compiledfilename += "_spy.py"
      codefiles.append(compiledfilename)      
      f = open(compiledfilename, "w")
      f.write(result)
      f.close()
      code += result
    elif fil.upper().endswith(".PY"):
      if not ok: raise
      __addsystemtypes(fcode,[],__state)
      codefiles.append(fil)
      code += open(fil).read() + '\n'
    else:
      if not ok: raise
      if fil.upper().endswith(".WEB"):
        f = open(fil)
        addsystemtypes(f.read().replace('\r\n','\n'),[],__state)
        f.close()
        code += "a = Spyder.core.parse(str(File(r\"%s\",Data).data().data()))\nif isinstance(a,tuple): a[0].show()\nelse: a.show()\n" % (os.path.abspath(fil))
        __addsystemtypes("File",[],__state)
        __addsystemtypes("Data",[],__state)
      else:
        format = "Data"
        p = fil.rfind(".")
        if p > -1:
          format += "_" + fil[p+1:].upper()
        code += "File(\"%s\",%s).data().show()\n" % (os.path.abspath(fil), format)
        __addsystemtypes("File",[],__state)
        __addsystemtypes(format,[],__state)

  __system.reset()
  
  __system.Spyder.core.__arraycount = __state.arraycount 
  __system.Spyder.core.register_spyderobject = __register_spyderobject
  __system.Spyder.core.__constructor = __constructor
  __system.Spyder.core.__set_temp = __set_temp
  __system.Spyder.core.__get_temp = __get_temp
  __system.Spyder.core.__clear_temp = __clear_temp
  __system.Spyder.core.exception = __exception  
  __system.Spyder.getvar = getvar
  __system.Spyder.getvars = getvars
  
  __system.changetempdir(__tempdir)        
  execode = __corecode
  allnam = [[v] + list(__state.allnamespaces[v]) for v in __state.allnamespaces]
  allnam.sort(__sortfunc)
  for nam, nr, namespace in allnam:
    a = [nam, namespace.codefiles, namespace.paths, namespace.imports, namespace.exports, namespace.defined]
    execode += __system.import_namespace(*a)      
  execode += code + '\n' 
  os.chdir(olddir)
  exec(execode, __shell)
  
def getvar(varname):
  if len(__system.var) == 0:
    init(None)
  if varname == "installdir":
    raise Exception("Value for $installdir is never fixed")
  elif varname not in __system.var:
    raise Exception("Variable %s is not defined in system.fly" % (varname))
  return __system.var[varname]

def getvars():
  if len(__system.var) == 0:
    init(None)
  return dict(__system.var)

def __export(spydertype):  
  types = __system.Spyder.core.__types__
  if spydertype not in types:
    __load_spydertype(spydertype)
  return types[spydertype]

def __module_export(spydermodule):  
  if spydermodule not in __shell["Spyder"].__dict__:
    __load_namespace(spydermodule)
  return __shell["Spyder"].__dict__[spydermodule]

class __wrapper(object):
  def __init__(self, wrapclassname):
    self.wrapclassname = wrapclassname
  def installed(self):
    return True    
  def __call__(self, *args, **kwargs):
    globals()["__export"](self.wrapclassname)
    self.wrapclass = globals()["__system"].Spyder.core.__types__[self.wrapclassname]
    return self.wrapclass(*args, **kwargs)  
  def help(self):
    globals()["__export"](self.wrapclassname)
    self.wrapclass = globals()["__system"].Spyder.core.__types__[self.wrapclassname]
    help(self.wrapclass)
  def __getattr__(self, *args, **kwargs):
    globals()["__export"](self.wrapclassname)
    self.wrapclass = globals()["__system"].Spyder.core.__types__[self.wrapclassname]
    if args == ("wrapclass",) and kwargs == {}: return self.wrapclass
    return getattr(self.wrapclass, *args, **kwargs)  

class __module_wrapper(object):
  def __init__(self, wrapmodulename):
    self.wrapmodulename = wrapmodulename
  def installed(self):
    return True    
  def help(self):
    self.wrapmodule = globals()["__module_export"](self.wrapmodulename)
  def __getattr__(self, *args, **kwargs):
    self.wrapmodule = globals()["__module_export"](self.wrapmodulename)
    return getattr(self.wrapmodule, *args, **kwargs)  
  def namespace(self):
    self.wrapmodule = globals()["__module_export"](self.wrapmodulename)
    return self.wrapmodule

class __class_stub:
  def __init__(self, classname, modulename):
    self.classname = classname
    self.modulename = modulename
  def installed(self):
    return False
  def __call__(self, *args, **kargs):
    if globals()["__state"] == None:
      init(None)      
    ret = globals()[self.classname]
    if ret is not self: return ret(*args, **kargs)    
    raise Exception("Class %s (module %s) is not installed" % (self.classname, self.modulename))
  def __getattr__(self, attr):
    if globals()["__state"] == None:
      init(None)
    ret = globals()[self.classname]
    if ret is not self: return getattr(ret, attr)      
    raise Exception("Class %s (module %s) is not installed" % (self.classname, self.modulename))

class __module_stub:
  def __init__(self, modulename):
    self.modulename = modulename
  def installed(self):
    return False    
  def __call__(self, *args, **kargs):
    if globals()["__state"] == None:
      init(None)      
    ret = globals()[self.modulename]
    if ret is not self: return ret(*args, **kargs)    
    raise Exception("Module %s is not installed" % (self.modulename,))
  def __getattr__(self, attr):
    if globals()["__state"] == None:
      init(None)
    ret = globals()[self.modulename]
    if ret is not self: return getattr(ret, attr)
    raise Exception("Module %s is not installed" % (self.modulename,))

__classes = [l.split() for l in open(__spyderdir+os.sep+"packages"+os.sep+"CLASSES.txt").readlines() if len(l.split()) == 2]
__modules = [l.split()[0] for l in open(__spyderdir+os.sep+"packages"+os.sep+"MODULES.txt").readlines() if len(l.split()) == 3]
for __c in __classes:
  for n in range(5):    #0-5x Array
    cname = __c[0]+n*"Array"
    globals()[cname] = __class_stub(cname, __c[1])
for __m in __modules:
  globals()[__m] = __module_stub(__m)